# -*- coding: utf-8 -*-
"""
QUaternion ESTimator as described by Shuster in :cite:p:`shuster1981` and
:cite:p:`shuster1978`.

We start to define the goal of finding an orthogonal matrix :math:`\\mathbf{A}`
that minimizes the loss function:

.. math::
    L(\\mathbf{A}) = \\frac{1}{2}\\sum_{i=1}^n a_i |\\hat{\\mathbf{W}}_i - \\mathbf{A}\\hat{\\mathbf{V}}_i|^2

where :math:`a_i` are a set of non-negative weights such that :math:`\\sum_{i=1}^na_i=1`,
:math:`\\hat{\\mathbf{V}}_i` are nonparallel **reference vectors**, and
:math:`\\hat{\\mathbf{W}}_i` are the corresponding **observation vectors**.

The gain function :math:`g(\\mathbf{A})` is defined by

.. math::
    g(\\mathbf{A}) = 1 - L(\\mathbf{A}) = \\sum_{i=1}^na_i\\,\\hat{\\mathbf{W}}_i^T\\mathbf{A}\\hat{\\mathbf{V}}_i

The loss function :math:`L(\\mathbf{A})` is at its minimum when the gain
function :math:`g(\\mathbf{A})` is at its maximum. The gain function can be
reformulated as:

.. math::
    g(\\mathbf{A}) = \\sum_{i=1}^na_i\\mathrm{tr}\\big(\\hat{\\mathbf{W}}_i^T\\mathbf{A}\\hat{\\mathbf{V}}_i\\big) = \\mathrm{tr}(\\mathbf{AB}^T)

where :math:`\\mathrm{tr}` is the `trace <https://en.wikipedia.org/wiki/Trace_(linear_algebra)>`_
of a matrix, and :math:`\\mathbf{B}` is the **attitude profile matrix**:

.. math::
    \\mathbf{B} = \\sum_{i=1}^na_i\\hat{\\mathbf{W}}_i\\hat{\\mathbf{V}}_i^T

The quaternion :math:`\\bar{\\mathbf{q}}` representing a rotation is defined by
Shuster as:

.. math::
    \\bar{\\mathbf{q}} = \\begin{bmatrix}\\mathbf{Q} \\\\ q\\end{bmatrix}
    = \\begin{bmatrix}\\hat{\\mathbf{X}}\\sin\\frac{\\theta}{2} \\\\ \\cos\\frac{\\theta}{2}\\end{bmatrix}

where :math:`\\hat{\\mathbf{X}}` is the axis of rotation, and :math:`\\theta`
is the angle of rotation about :math:`\\hat{\\mathbf{X}}`.

.. warning::
    The definition of a quaternion used by Shuster sets the vector part
    :math:`\\mathbf{Q}` followed by the scalar part :math:`q`. This module,
    however, will return the estimated quaternion with the *scalar part first*
    and followed by the vector part: :math:`\\bar{\\mathbf{q}} = \\begin{bmatrix}q
    & \\mathbf{Q}\\end{bmatrix}`

Because the quaternion works as a versor, it must satisfy:

.. math::
    \\bar{\\mathbf{q}}^T\\bar{\\mathbf{q}} = |\\mathbf{Q}|^2 + q^2 = 1

The attitude matrix :math:`\\mathbf{A}` is related to the quaternion by:

.. math::
    \\mathbf{A}(\\bar{\\mathbf{q}}) = (q^2-\\mathbf{Q}\\cdot\\mathbf{Q})\\mathbf{I} + 2\\mathbf{QQ}^T + 2q\\lfloor\\mathbf{Q}\\rfloor_\\times

where :math:`\\mathbf{I}` is the identity matrix, and :math:`\\lfloor\\mathbf{Q}\\rfloor_\\times`
is the **antisymmetric matrix** of :math:`\\mathbf{Q}`, a.k.a. the
`skew-symmetric matrix <https://en.wikipedia.org/wiki/Skew-symmetric_matrix>`_:

.. math::
    \\lfloor\\mathbf{Q}\\rfloor_\\times =
    \\begin{bmatrix}0 & Q_3 & -Q_2 \\\\ -Q_3 & 0 & Q_1 \\\\ Q_2 & -Q_1 & 0\\end{bmatrix}

Now the gain function can be rewritten again, but in terms of quaternions:

.. math::
    g(\\bar{\\mathbf{q}}) = (q^2-\\mathbf{Q}\\cdot\\mathbf{Q})\\mathrm{tr}\\mathbf{B}^T + 2\\mathrm{tr}\\big(\\mathbf{QQ}^T\\mathbf{B}^T\\big) + 2q\\mathrm{tr}\\big(\\lfloor\\mathbf{Q}\\rfloor_\\times\\mathbf{B}^T\\big)

A further simplification gives:

.. math::
    g(\\bar{\\mathbf{q}}) = \\bar{\\mathbf{q}}^T\\mathbf{K}\\bar{\\mathbf{q}}

where the :math:`4\\times 4` matrix :math:`\\mathbf{K}` is given by:

.. math::
    \\mathbf{K} = \\begin{bmatrix} \\mathbf{S} - \\sigma\\mathbf{I} & \\mathbf{Z} \\\\ \\mathbf{Z}^T & \\sigma \\end{bmatrix}

using the helper values:

.. math::
    \\begin{array}{rcl}
    \\sigma &=& \\mathrm{tr}\\mathbf{B} \\\\ && \\\\
    \\mathbf{S} &=& \\mathbf{B} + \\mathbf{B}^T \\\\ && \\\\
    \\mathbf{Z} &=& \\sum_{i=1}^na_i\\big(\\hat{\\mathbf{W}}_i\\times\\hat{\\mathbf{V}}_i\\big)
    \\end{array}

.. note::
    :math:`\\mathbf{Z}` can be also defined from :math:`\\lfloor\\mathbf{Z}\\rfloor_\\times = \\mathbf{B} - \\mathbf{B}^T`

A new gain function :math:`g'(\\bar{\\mathbf{q}})` with `Lagrange multipliers
<https://en.wikipedia.org/wiki/Lagrange_multiplier>`_ is defined:

.. math::
    g'(\\bar{\\mathbf{q}}) = \\bar{\\mathbf{q}}^T\\mathbf{K}\\bar{\\mathbf{q}} - \\lambda\\bar{\\mathbf{q}}^T\\bar{\\mathbf{q}}

It is verified that :math:`\\mathbf{K}\\bar{\\mathbf{q}}=\\lambda\\bar{\\mathbf{q}}`.
Thus, :math:`g(\\bar{\\mathbf{q}})` will be maximized if :math:`\\bar{\\mathbf{q}}_\\mathrm{opt}`
is chosen to be the eigenvector of :math:`\\mathbf{K}` belonging to the largest
eigenvalue of :math:`\\mathbf{K}`:

.. math::
    \\mathbf{K}\\bar{\\mathbf{q}}_\\mathrm{opt} = \\lambda_\\mathrm{max}\\bar{\\mathbf{q}}_\\mathrm{opt}

which is the desired result. This equation can be rearranged to read, for any
eigenvalue :math:`\\lambda`:

.. math::
    \\lambda = \\sigma + \\mathbf{Z}\\cdot\\mathbf{Y}

where :math:`\\mathbf{Y}` is the `Gibbs vector
<https://en.wikipedia.org/wiki/Rotation_formalisms_in_three_dimensions#Rodrigues_vector>`_,
a.k.a. the **Rodrigues vector**, defined as:

.. math::
    \\mathbf{Y} = \\frac{\\mathbf{Q}}{q} = \\hat{\\mathbf{X}}\\tan\\frac{\\theta}{2}

rewriting the quaternion as:

.. math::
    \\bar{\\mathbf{q}} = \\frac{1}{\\sqrt{1+|\\mathbf{Y}|^2}} = \\begin{bmatrix}\\mathbf{Y}\\\\ 1 \\end{bmatrix}

:math:`\\mathbf{Y}` and :math:`\\bar{\\mathbf{q}}` are representations of the
optimal attitude solution when :math:`\\lambda` is equal to :math:`\\lambda_\\mathrm{max}`,
leading to an equation for the eigenvalues:

.. math::
    \\lambda = \\sigma + \\mathbf{Z}^T \\frac{1}{(\\lambda+\\sigma)\\mathbf{I}-\\mathbf{S}}\\mathbf{Z}

which is equivalent to the characteristic equation of the eigenvalues of :math:`\\mathbf{K}`

With the aid of `Cayley-Hamilton theorem <https://en.wikipedia.org/wiki/Cayley%E2%80%93Hamilton_theorem>`_
we can get rid of the Gibbs vector to find a more convenient expression of the
characteristic equation:

.. math::
    \\lambda^4-(a+b)\\lambda^2-c\\lambda+(ab+c\\sigma-d)=0

where:

.. math::
    \\begin{array}{rcl}
    a &=& \\sigma^2-\\kappa \\\\ && \\\\
    b &=& \\sigma^2 + \\mathbf{Z}^T\\mathbf{Z} \\\\ && \\\\
    c &=& \\Delta + \\mathbf{Z}^T\\mathbf{SZ} \\\\ && \\\\
    d &=& \\mathbf{Z}^T\\mathbf{S}^2\\mathbf{Z} \\\\ && \\\\
    \\sigma &=& \\frac{1}{2}\\mathrm{tr}\\mathbf{S} \\\\ && \\\\
    \\kappa &=& \\mathrm{tr}\\big(\\mathrm{adj}(\\mathbf{S})\\big) \\\\ && \\\\
    \\Delta &=& \\mathrm{det}(\\mathbf{S})
    \\end{array}

To find :math:`\\lambda` we implement the `Newton-Raphson method
<https://en.wikipedia.org/wiki/Newton%27s_method>`_ using the sum of the
weights :math:`a_i` (in the beginning is constrained to be equal to 1) as a
starting value.

.. math::
    \\lambda_{t+1} \\gets \\lambda_t - \\frac{f(\\lambda)}{f'(\\lambda)}
    = \\lambda_t - \\frac{\\lambda^4-(a+b)\\lambda^2-c\\lambda+(ab+c\\sigma-d)}{4\\lambda^3-2(a+b)\\lambda-c}

For sensor accuracies better than 1 arc-min (1 degree) the accuracy of a 64-bit
word is exhausted after only one iteration.

Finally, the **optimal quaternion** describing the attitude is found as:

.. math::
    \\bar{\\mathbf{q}}_\\mathrm{opt} = \\frac{1}{\\sqrt{\\gamma^2+|\\mathbf{X}|^2}} \\begin{bmatrix}\\mathbf{X}\\\\ \\gamma \\end{bmatrix}

with:

.. math::
    \\begin{array}{rcl}
    \\mathbf{X} &=& (\\alpha\\mathbf{I} + \\beta\\mathbf{S} + \\mathbf{S}^2)\\mathbf{Z} \\\\ && \\\\
    \\gamma &=& (\\lambda + \\sigma)\\alpha - \\Delta \\\\ && \\\\
    \\alpha &=& \\lambda^2 - \\sigma^2 + \\kappa \\\\ && \\\\
    \\beta &=& \\lambda - \\sigma
    \\end{array}

This solution can still lead to an indeterminant result if both :math:`\\gamma`
and :math:`\\mathbf{X}` vanish simultaneously. :math:`\\gamma` vanishes if and
only if the angle of rotation is equal to :math:`\\pi`, even if
:math:`\\mathbf{X}` does not vanish along.

"""

from typing import Union
import numpy as np

from ..common.mathfuncs import cosd
from ..common.mathfuncs import sind
from ..common.constants import MUNICH_LATITUDE
from ..common.constants import MUNICH_LONGITUDE
from ..common.constants import MUNICH_HEIGHT
from ..utils.core import _assert_numerical_iterable
from ..utils.core import _assert_acc_mag_inputs

# Reference Observations in Munich, Germany
from ..utils.wmm import WMM
from ..utils.wgs84 import WGS
GRAVITY = WGS().normal_gravity(MUNICH_LATITUDE, MUNICH_HEIGHT)
wmm = WMM(latitude=MUNICH_LATITUDE, longitude=MUNICH_LONGITUDE, height=MUNICH_HEIGHT)
REFERENCE_MAGNETIC_VECTOR = np.array([wmm.X, wmm.Y, wmm.Z])

def _set_magnetic_field_vector(magnetic_dip: Union[int, float, list, tuple, np.ndarray]):
    """
    Set the magnetic reference vector.

    Parameters
    ----------
    magnetic_dip : float, array-like
        Magnetic dip, in degrees, or local geomagnetic reference as a
        three-dimensional vector.

    """
    if isinstance(magnetic_dip, bool):
        raise TypeError("magnetic_dip must be given as a float, list, tuple or NumPy array. Got bool")
    elif isinstance(magnetic_dip, (float, int)):
        magnetic_field = np.array([cosd(magnetic_dip), 0., sind(magnetic_dip)])
    elif isinstance(magnetic_dip, (list, tuple, np.ndarray)):
        if not all(isinstance(x, (float, int)) for x in magnetic_dip):
            raise TypeError("magnetic_dip must be an array of floats. Contains non-numeric values.")
        magnetic_field = np.copy(magnetic_dip)
    elif magnetic_dip is None:
        magnetic_field = REFERENCE_MAGNETIC_VECTOR
    else:
        raise TypeError(f"magnetic_dip must be given as a float, list, tuple or NumPy array. Got {type(magnetic_dip)}")
    if magnetic_field.shape != (3,):
        raise ValueError(f"magnetic_dip array must contain 3 elements. Got {magnetic_field.shape}")
    return magnetic_field / np.linalg.norm(magnetic_field)

class QUEST:
    """
    QUaternion ESTimator

    Parameters
    ----------
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration in in m/s^2
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field in mT
    weights : array-like
        Array with two weights. One per sensor measurement.
    magnetic_dip : float
        Local magnetic inclination angle, in degrees.
    gravity : float
        Local normal gravity, in m/s^2.

    Attributes
    ----------
    acc : numpy.ndarray
        N-by-3 array with N accelerometer samples.
    mag : numpy.ndarray
        N-by-3 array with N magnetometer samples.
    w : numpy.ndarray
        Weights for each observation.

    Raises
    ------
    ValueError
        When dimension of input arrays ``acc`` and ``mag`` are not equal.

    Examples
    --------
    >>> acc_data.shape, mag_data.shape      # NumPy arrays with sensor data
    ((1000, 3), (1000, 3))
    >>> from ahrs.filters import QUEST
    >>> orientation = QUEST(acc=acc_data, mag=mag_data)
    >>> orientation.Q.shape                 # Estimated attitudes as Quaternions
    (1000, 4)
    >>> orientation.Q
    array([[-0.09867706, -0.33683592, -0.52706394, -0.77395607],
           [-0.10247491, -0.33710813, -0.52117549, -0.77732433],
           [-0.10082646, -0.33658091, -0.52082828, -0.77800078],
           ...,
           [-0.78760687, -0.57789515,  0.2131519,  -0.01669966],
           [-0.78683706, -0.57879487,  0.21313092, -0.02142776],
           [-0.77869223, -0.58616905,  0.22344478, -0.01080235]])

    """
    def __init__(self, acc: np.ndarray = None, mag: np.ndarray = None, **kw):
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.w: np.ndarray = kw.get('weights', 0.5*np.ones(2))
        # Reference measurements
        self.m_q: np.ndarray = _set_magnetic_field_vector(kw.get('magnetic_dip'))
        self.g_q: np.ndarray = np.array([0., 0., 1.])  # Normal Gravity vector
        if self.acc is not None and self.mag is not None:
            self.Q: np.ndarray = self._compute_all()

    def _compute_all(self) -> np.ndarray:
        """
        Estimate the quaternions given all data.

        Attributes ``acc`` and ``mag`` must contain data.

        Returns
        -------
        Q : numpy.ndarray
            M-by-4 Array with all estimated quaternions, where M is the number
            of samples.

        """
        _assert_acc_mag_inputs(self.acc, self.mag)
        acc, mag = np.copy(self.acc), np.copy(self.mag)
        if acc.ndim < 2:
            return self.estimate(acc, mag)
        num_samples = len(acc)
        return np.array([self.estimate(acc[t], mag[t]) for t in range(num_samples)])

    def estimate(self, acc: np.ndarray, mag: np.ndarray) -> np.ndarray:
        """
        Attitude Estimation.

        Parameters
        ----------
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer in T

        Returns
        -------
        q : numpy.ndarray
            Estimated attitude as a quaternion.

        """
        _assert_numerical_iterable(acc, 'Gravitational acceleration vector')
        _assert_numerical_iterable(mag, 'Geomagnetic field vector')
        acc = acc/np.linalg.norm(acc)
        mag = mag/np.linalg.norm(mag)
        B = self.w[0]*np.outer(acc, self.g_q) + self.w[1]*np.outer(mag, self.m_q)   # Attitude profile matrix
        S = B + B.T
        z = np.array([B[1, 2]-B[2, 1], B[2, 0]-B[0, 2], B[0, 1]-B[1, 0]])   # Pseudovector (Axial vector)
        # Parameters of characeristic equation (eq. 63)
        sigma = B.trace()
        Delta = np.linalg.det(S)
        adjS = Delta*np.linalg.inv(S)
        kappa = adjS.trace()
        ### Elements of characteristic polynomial (eq. 71)
        a = sigma**2 - kappa
        b = sigma**2 + z.T @ z
        c = Delta + z.T @ S @ z
        d = z.T @ S@S @ z
        # Newton-Raphson method (eq. 70)
        k = a*b + c*sigma - d
        l_max = l_old = self.w.sum()
        num_iters = 0
        while abs(l_old-l_max) > 1e-8 or num_iters <= 10:
            l_old = l_max
            phi = l_max**4 - (a+b)*l_max**2 - c*l_max + k
            phi_prime = 4.0*l_max**3 - 2.0*(a+b)*l_max - c
            l_max -= phi/phi_prime
            num_iters += 1
        # (eq. 66)
        alpha = l_max**2 - sigma**2 + kappa
        beta = l_max - sigma
        gamma = (l_max + sigma)*alpha - Delta
        Chi = (alpha*np.identity(3) + beta*S + S@S)@ z      # (eq. 68)
        # Optimal Quaternion (eq. 69)
        q = [gamma, *Chi]
        q /= np.linalg.norm(q)
        return q



# -*- coding: utf-8 -*-
"""
.. contents:: Table of Contents
    :local:
    :depth: 2

This is an orientation filter applicable to IMUs consisting of tri-axial
gyroscopes and accelerometers, and MARG arrays, which also include tri-axial
magnetometers, proposed by Sebastian Madgwick :cite:p:`madgwick2010`.

The filter employs a quaternion representation of orientation to describe the
nature of orientations in three-dimensions and is not subject to the
singularities associated with an Euler angle representation, allowing
accelerometer and magnetometer data to be used in an analytically derived and
optimised gradient-descent algorithm to compute the direction of the gyroscope
measurement error as a quaternion derivative.

Innovative aspects of this filter include:

- A single adjustable parameter defined by observable systems characteristics.
- An analytically derived and optimised gradient-descent algorithm enabling
  performance at low sampling rates.
- On-line magnetic distortion compensation algorithm.
- Gyroscope bias drift compensation.

Rewritten in Python from the `original implementation <https://x-io.co.uk/open-source-imu-and-ahrs-algorithms/>`_
conceived by `Sebastian Madgwick <https://scholar.google.co.uk/citations?user=pVoC0_cAAAAJ>`_.

Orientation from angular rate
-----------------------------

The orientation of the Earth frame relative to the sensor frame
:math:`\\mathbf{q}_{\\omega, t}=\\begin{bmatrix}q_w & q_x & q_y & q_z\\end{bmatrix}`
at time :math:`t` can be computed by numerically integrating the quaternion
derivative :math:`\\dot{\\mathbf{q}}_t=\\frac{1}{2}\\mathbf{q}_{t-1}\\mathbf{\\omega}_t` as:

.. math::
    \\begin{array}{rcl}
    \\mathbf{q}_{\\omega, t} &=& \\,\\mathbf{q}_{t-1} + \\,\\dot{\\mathbf{q}}_{\\omega, t}\\Delta t\\\\
    &=& \\,\\mathbf{q}_{t-1} + \\frac{1}{2}\\big(\\,\\mathbf{q}_{t-1}\\mathbf{\\,^S\\omega_t}\\big)\\Delta t
    \\end{array}

where :math:`\\Delta t` is the sampling period and :math:`^S\\omega=\\begin{bmatrix}0 & \\omega_x & \\omega_y & \\omega_z\\end{bmatrix}`
is the tri-axial angular rate, in rad/s, measured in the sensor frame and
represented as a pure quaternion.

.. note::
    The multiplication of quaternions (included pure quaternions) is performed
    as a `Hamilton product <https://en.wikipedia.org/wiki/Quaternion#Hamilton_product>`_.
    All quaternion products explained here follow this procedure. For further
    details on how to compute it, see the `quaternion <../quaternion.html>`_
    documentation page.

The sub-script :math:`\\omega` in :math:`\\mathbf{q}_\\omega` indicates that it
is calculated from angular rates.

A more detailed explanation of the orientation estimation solely based on
angular rate can be found in the documentation of the `AngularRate
<./angular.html>`_ estimator.

Orientation as solution of Gradient Descent
-------------------------------------------

A quaternion representation requires a complete solution to be found. This may
be achieved through the formulation of an optimization problem where an
orientation of the sensor, :math:`\\mathbf{q}`, is that which aligns any
*predefined reference* in the Earth frame,
:math:`^E\\mathbf{d}=\\begin{bmatrix}0 & d_x & d_y & d_z\\end{bmatrix}`, with
its corresponding *measured* direction in the sensor frame,
:math:`^S\\mathbf{s}=\\begin{bmatrix}0 & s_x & s_y & s_z\\end{bmatrix}`.

Thus, the `objective function <https://en.wikipedia.org/wiki/Loss_function>`_
is:

.. math::
    \\begin{array}{rcl}
    f( \\mathbf{q}, \\,^E\\mathbf{d}, \\,^S\\mathbf{s}) &=&  \\mathbf{q}^*\\,^E\\mathbf{d} \\,\\mathbf{q}-\\,^S\\mathbf{s} \\\\
    &=&\\begin{bmatrix}
    2d_x(\\frac{1}{2}-q_y^2-q_z^2) + 2d_y(q_wq_z+q_xq_y) + 2d_z(q_xq_z-q_wq_y) - s_x \\\\
    2d_x(q_xq_y-q_wq_z) + 2d_y(\\frac{1}{2}-q_x^2-q_z^2) + 2d_z(q_wq_x+q_yq_z) - s_y \\\\
    2d_x(q_wq_y+q_xq_z) + 2d_y(q_yq_z-q_wq_x) + 2d_z(\\frac{1}{2}-q_x^2-q_y^2) - s_z
    \\end{bmatrix}
    \\end{array}

where :math:`\\mathbf{q}^*` is the `conjugate <https://mathworld.wolfram.com/QuaternionConjugate.html>`_
of :math:`\\mathbf{q}`. Consequently, :math:`\\mathbf{q}` is found as
the solution to:

.. math::
    \\mathrm{min}\\; f( \\mathbf{q}, \\,^E\\mathbf{d}, \\,^S\\mathbf{s})

The suggested approach of this estimator is to use the `Gradient Descent
Algorithm <https://en.wikipedia.org/wiki/Gradient_descent>`_ to compute the
solution.

From an *initial guess* :math:`\\mathbf{q}_0` and a step-size :math:`\\mu`,
the GDA for :math:`n` iterations, which estimates :math:`n+1` orientations, is
described as:

.. math::
     \\mathbf{q}_{k+1} =  \\mathbf{q}_k-\\mu\\frac{\\nabla f( \\mathbf{q}_k, \\,^E\\mathbf{d}, \\,^S\\mathbf{s})}{\\|\\nabla f( \\mathbf{q}_k, \\,^E\\mathbf{d}, \\,^S\\mathbf{s})\\|}

where :math:`k=0,1,2\\dots n`, and the `gradient <https://en.wikipedia.org/wiki/Gradient>`_
of the solution is defined by the objective function and its Jacobian:

.. math::
    \\nabla f( \\mathbf{q}_k, \\,^E\\mathbf{d}, \\,^S\\mathbf{s}) = J( \\mathbf{q}_k, \\,^E\\mathbf{d})^T f( \\mathbf{q}_k, \\,^E\\mathbf{d}, \\,^S\\mathbf{s})

The `Jacobian <https://en.wikipedia.org/wiki/Jacobian_matrix_and_determinant>`_
of the objective function is:

.. math::
    \\begin{array}{rcl}
    J( \\mathbf{q}_k, \\,^E\\mathbf{d}) &=&\\begin{bmatrix}
    \\frac{\\partial f( \\mathbf{q}, \\,^E\\mathbf{d}, \\,^S\\mathbf{s})}{\\partial q_w} &
    \\frac{\\partial f( \\mathbf{q}, \\,^E\\mathbf{d}, \\,^S\\mathbf{s})}{\\partial q_x} &
    \\frac{\\partial f( \\mathbf{q}, \\,^E\\mathbf{d}, \\,^S\\mathbf{s})}{\\partial q_y} &
    \\frac{\\partial f( \\mathbf{q}, \\,^E\\mathbf{d}, \\,^S\\mathbf{s})}{\\partial q_z} &
    \\end{bmatrix}\\\\
    &=& \\begin{bmatrix}
    2d_yq_z-2d_zq_y  & 2d_yq_y+2d_zq_z         & -4d_xq_y+2d_yq_x-2d_zq_w & -4d_xq_z+2d_yq_w+2d_zq_x \\\\
    -2d_xq_z+2d_zq_x & 2d_xq_y-4d_yq_x+2d_zq_w & 2d_xq_x+2d_zq_z          & -2d_xq_w-4d_yq_z+2d_zq_y \\\\
    2d_xq_y-2d_yq_x  & 2d_xq_z-2d_yq_w-4d_zq_x & 2d_xq_w+2d_yq_z-4d_zq_y  & 2d_xq_x+2d_yq_y
    \\end{bmatrix}
    \\end{array}

This general form of the algorithm can be applied to a field predefined in any
direction, as it will be shown for IMU and MARG systems.

The gradient quaternion :math:`\\mathbf{q}_{\\nabla, t}`, computed at time
:math:`t`, is based on a previous estimation :math:`\\mathbf{q}_{t-1}` and
the *objective function gradient* :math:`\\nabla f`:

.. math::
     \\mathbf{q}_{\\nabla, t} =  \\mathbf{q}_{t-1}-\\mu_t\\frac{\\nabla f}{\\|\\nabla f\\|}

An optimal value of :math:`\\mu_t` ensures that the convergence rate of
:math:`\\mathbf{q}_{\\nabla, t}` is limited to the physical orientation
rate. It can be calculated with:

.. math::
    \\mu_t = \\alpha\\|\\,\\dot{\\mathbf{q}}_{\\omega, t}\\|\\Delta t

with :math:`\\dot{\\mathbf{q}}_{\\omega, t}` being the physical orientation
rate measured by the gyroscopes, and :math:`\\alpha>1` is an augmentation of
:math:`\\mu` dealing with the noise of accelerometers and magnetometers.

An estimated orientation of the sensor frame relative to the earth frame,
:math:`\\mathbf{q}_t`, is obtained through the *weighted fusion of the
orientation calculations*, :math:`\\mathbf{q}_{\\omega, t}` and :math:`\\mathbf{q}_{\\nabla, t}`
with a simple **complementary filter**:

.. math::
     \\mathbf{q}_t = \\gamma_t  \\mathbf{q}_{\\nabla, t} + (1-\\gamma_t) \\mathbf{q}_{\\omega, t}

where :math:`\\gamma_t` and :math:`(1-\\gamma_t)` are the weights, ranging
between 0 and 1, applied to each orientation calculation. An optimal value of
:math:`\\gamma_t` ensures that the weighted divergence of :math:`\\mathbf{q}_{\\omega, t}`
is equal to the weighted convergence of :math:`\\mathbf{q}_{\\nabla, t}`.
This is expressed with:

.. math::
    (1-\\gamma_t)\\beta = \\gamma_t\\frac{\\mu_t}{\\Delta t}

defining :math:`\\frac{\\mu_t}{\\Delta t}` as the *convergence rate* of
:math:`\\mathbf{q}_\\nabla`, and :math:`\\beta` as the *divergence rate* of
:math:`\\mathbf{q}_\\omega` expressed as the magnitude of a quaternion
derivative corresponding to the gyroscope measurement error.

If :math:`\\alpha` is very large then :math:`\\mu` becomes very large making
:math:`\\mathbf{q}_{t-1}` negligible in the **objective function gradient**
simplifying it to the approximation:

.. math::
     \\mathbf{q}_{\\nabla, t} \\approx -\\mu_t\\frac{\\nabla f}{\\|\\nabla f\\|}

This also simplifies the relation of :math:`\\gamma` and :math:`\\beta`:

.. math::
    \\gamma \\approx \\frac{\\beta\\Delta t}{\\mu_t}

which further reduces the estimation to:

.. math::
    \\begin{array}{rcl}
     \\mathbf{q}_t &=&  \\mathbf{q}_{t-1} +  \\dot{\\mathbf{q}}_t\\Delta t \\\\
    &=&  \\mathbf{q}_{t-1} + \\big( \\dot{\\mathbf{q}}_{\\omega, t} - \\beta \\dot{\\mathbf{q}}_{\\epsilon, t}\\big)\\Delta t \\\\
    &=&  \\mathbf{q}_{t-1} + \\big( \\dot{\\mathbf{q}}_{\\omega, t} - \\beta\\frac{\\nabla f}{\\|\\nabla f\\|}\\big)\\Delta t
    \\end{array}

where :math:`\\dot{\\mathbf{q}}_t` is the **estimated rate of change of
orienation** defined by :math:`\\beta` and its direction error
:math:`\\dot{\\mathbf{q}}_{\\epsilon, t}=\\frac{\\nabla f}{\\|\\nabla f\\|}`.

In summary, the filter calculates the orientation :math:`\\mathbf{q}_{t}`
by numerically integrating the estimated orientation rate :math:`\\dot{\\mathbf{q}}_t`.
It computes :math:`\\dot{\\mathbf{q}}_t` as the rate of change of
orientation measured by the gyroscopes, :math:`\\dot{\\mathbf{q}}_{\\omega, t}`,
with the magnitude of the gyroscope measurement error, :math:`\\beta`, removed
in the direction of the estimated error, :math:`\\dot{\\mathbf{q}}_{\\epsilon, t}`,
computed from accelerometer and magnetometer measurements.

Orientation from IMU
--------------------

Two main geodetic properties can be used to build Earth's reference:

- The `gravitational force <https://en.wikipedia.org/wiki/Gravity_of_Earth>`_
  :math:`^E\\mathbf{g}` represented as a vector and measured with a tri-axial
  accelerometer.
- The `geomagnetic field <https://en.wikipedia.org/wiki/Earth%27s_magnetic_field>`_
  :math:`^E\\mathbf{b}` represented as a vector and measured with a tri-axial
  magnetometer.

Earth's shape is not uniform and a geographical location is usually provided to
obtain the references' true values. Madgwick's filter, however, uses normalized
references, making their magnitudes, irrespective of their location, always
equal to 1.

.. note::
    All real vectors in a three-dimensional euclidean operating with
    quaternions will be considered pure quaternions. That is, given a
    three-dimensional vector :math:`\\mathbf{x}=\\begin{bmatrix}x&y&z\\end{bmatrix}\\in\\mathbb{R}^3`,
    it will be redefined as :math:`\\mathbf{x}=\\begin{bmatrix}0&x&y&z\\end{bmatrix}\\in\\mathbf{H}^4`.

To obtain the objective function of the **gravitational acceleration**, we
assume, by convention, that the vertical Z-axis is defined by the direction of
the gravity :math:`^E\\mathbf{g}=\\begin{bmatrix}0 & 0 & 0 & 1\\end{bmatrix}`.

Substituting :math:`^E\\mathbf{g}` and the *normalized* accelerometer
measurement :math:`^S\\mathbf{a}=\\begin{bmatrix}0 & a_x & a_y & a_z\\end{bmatrix}`
for :math:`^E\\mathbf{d}` and :math:`^S\\mathbf{s}`, respectively, yields a new
objective function and its Jacobian particular to the acceleration:

.. math::
    \\begin{array}{c}
    f_g( \\mathbf{q}, \\,^S\\mathbf{a}) = \\begin{bmatrix}
    2(q_xq_z-q_wq_y)-a_x \\\\ 2(q_wq_x+q_yq_z)-a_y \\\\ 2(\\frac{1}{2}-q_x^2-q_y^2)-a_z
    \\end{bmatrix} \\\\ \\\\
    J_g( \\mathbf{q})=\\begin{bmatrix}
    -2q_y & 2q_z & -2q_w & 2q_x \\\\
    2q_x & 2q_w & 2q_z & 2q_y \\\\
    0 & -4q_x & -4q_y & 0
    \\end{bmatrix}
    \\end{array}

The function gradient is defined by the sensor measurements at time :math:`t`:

.. math::
    \\nabla f = J_g^T( \\mathbf{q}_{t-1})f_g( \\mathbf{q}_{t-1}, \\,^S\\mathbf{a}_t)

So, the estimation of the orientation using inertial sensors only (gyroscopes
and accelerometers) becomes:

.. math::
    \\begin{array}{rcl}
     \\mathbf{q}_t &=&  \\mathbf{q}_{t-1} +  \\dot{\\mathbf{q}}_t\\Delta t \\\\
    &=&  \\mathbf{q}_{t-1} + \\Big( \\dot{\\mathbf{q}}_{\\omega, t} - \\beta\\frac{\\nabla f}{\\|\\nabla f\\|}\\Big) \\Delta t \\\\
    &=&  \\mathbf{q}_{t-1} + \\Big( \\dot{\\mathbf{q}}_{\\omega, t} - \\beta\\frac{J_g^T( \\mathbf{q}_{t-1})f_g( \\mathbf{q}_{t-1}, \\,^S\\mathbf{a}_t)}{\\|J_g^T( \\mathbf{q}_{t-1})f_g( \\mathbf{q}_{t-1}, \\,^S\\mathbf{a}_t)\\|}\\Big) \\Delta t
    \\end{array}

Orientation from MARG
---------------------

The gravity and the angular velocity are good parameters for an estimation over
a short period of time. But they don't hold for longer periods of time,
especially estimating the heading orientation of the system, as the gyroscope
measurements, prone to drift, are instantaneous and local, while the
accelerometer computes the roll and pitch orientations only.

Therefore, it is always very convenient to add a reference that provides
constant information about the heading angle (a.k.a. yaw). Earth's magnetic
field is usually the chosen reference, as it fairly keeps a constant reference [#]_.

The mix of Magnetic, Angular Rate and Gravity (MARG) is the most prevalent
solution in the majority of attitude estimation systems.

The reference magnetic field :math:`^E\\mathbf{b}=\\begin{bmatrix}0 & b_x & b_y & b_z\\end{bmatrix}`
in Earth's frame, has components along the three axes of `NED coordinates
<https://en.wikipedia.org/wiki/Local_tangent_plane_coordinates#Local_north,_east,_down_(NED)_coordinates>`_
(North-East-Down), which can be obtained using the `World Magnetic Model
<https://www.ngdc.noaa.gov/geomag/WMM/>`_.

Madgwick's estimator, nonetheless, assumes the East component of the magnetic
field (along Y-axis) is negligible, further reducing the reference magnetic
vector to:

.. math::
    \\mathbf{b}=\\begin{bmatrix}0 & b_x & 0 & b_z\\end{bmatrix}

The *measured* direction of Earth's magnetic field in the Earth frame at time
:math:`t`, :math:`^E\\mathbf{h}_t`, can be computed as the **normalized**
magnetometer measurement, :math:`^S\\mathbf{m}_t`, rotated by the orientation
of the sensor computed in the previous estimation, :math:`\\mathbf{q}_{t-1}`.

.. math::
    ^E\\mathbf{h}_t = \\begin{bmatrix}0 & h_x & h_y & h_z\\end{bmatrix} =
    \\,\\mathbf{q}_{t-1}\\,^S\\mathbf{m}_t\\,\\mathbf{q}_{t-1}^*

The effect of an erroneous inclination of the measured direction Earth's
magnetic field, :math:`^E\\mathbf{h}_t`, can be corrected if the filter's
reference direction of the geomagnetic field, :math:`^E\\mathbf{b}_t`, is of
the same inclination. This is achieved by computing :math:`^E\\mathbf{b}_t` as
a normalized :math:`^E\\mathbf{h}_t` to have only components in X- and Z-axes
of the Earth frame.

.. math::
    ^E\\mathbf{b}_t = \\begin{bmatrix}0 & \\sqrt{h_x^2+h_y^2} & 0 & h_z\\end{bmatrix}

This way ensures that magnetic disturbances are limited to only affect the
estimated heading component of orientation. It also eliminates the need for the
reference direction of the Earth's magnetic field to be predefined.

Substituting :math:`^E\\mathbf{b}` and the normalized magnetometer normalized
:math:`^S\\mathbf{m}` to form the *objective function* and *Jacobian* we get:

.. math::
    \\begin{array}{c}
    f_b( \\mathbf{q}, \\,^E\\mathbf{b}, \\,^S\\mathbf{m}) = \\begin{bmatrix}
    2b_x(\\frac{1}{2}-q_y^2-q_z^2) + 2b_z(q_xq_z-q_wq_y)-m_x \\\\
    2b_x(q_xq_y-q_wq_z) + 2b_z(q_wq_x+q_yq_z)-m_y \\\\
    2b_x(q_wq_y+q_xq_z) + 2b_z(\\frac{1}{2}-q_x^2-q_y^2)-m_z
    \\end{bmatrix} \\\\ \\\\
    J_b( \\mathbf{q}, \\,^E\\mathbf{b})=\\begin{bmatrix}
    -2b_zq_y          & 2b_zq_z         & -4b_xq_y-2b_zq_w & -4b_xq_z+2b_zq_x \\\\
    -2b_xq_z+2b_zq_x  & 2b_xq_y+2b_zq_w & 2b_xq_x+2b_zq_z  & -2b_xq_w+2b_zq_y \\\\
    2b_xq_y           & 2b_xq_z-4b_zq_x & 2b_xq_w-4b_zq_y  & 2b_xq_x
    \\end{bmatrix}
    \\end{array}

The measurements and reference directions of both fields, gravity and magnetic
field, are combined, where the solution surface has a minimum defined by a
single point, as long as the northerly magnetic intensity is defined (:math:`b_x\\neq 0`):

.. math::
    \\begin{array}{c}
    f_{g,b}( \\mathbf{q}, \\,^S\\mathbf{a}, \\,^E\\mathbf{b}, \\,^S\\mathbf{m})=
    \\begin{bmatrix}f_g( \\mathbf{q}, \\,^S\\mathbf{a}) \\\\ f_b( \\mathbf{q}, \\,^E\\mathbf{b}, \\,^S\\mathbf{m})\\end{bmatrix}\\\\ \\\\
    J_{g,b}( \\mathbf{q}, \\,^E\\mathbf{b})=
    \\begin{bmatrix}J_g^T( \\mathbf{q}) \\\\ J_b^T( \\mathbf{q}, \\,^E\\mathbf{b})\\end{bmatrix}
    \\end{array}

Simliar to the implementation with IMU, the estimation of the new quaternion
will be:

.. math::
     \\mathbf{q}_t =  \\mathbf{q}_{t-1} + \\Big( \\dot{\\mathbf{q}}_{\\omega, t} - \\beta\\frac{J_{g,b}^T( \\mathbf{q}_{t-1}, \\,^E\\mathbf{b})f_{g,b}( \\mathbf{q}_{t-1}, \\,^S\\mathbf{a}, \\,^E\\mathbf{b}, \\,^S\\mathbf{m})}{\\|J_{g,b}^T( \\mathbf{q}_{t-1}, \\,^E\\mathbf{b})f_{g,b}( \\mathbf{q}_{t-1}, \\,^S\\mathbf{a}, \\,^E\\mathbf{b}, \\,^S\\mathbf{m})\\|}\\Big) \\Delta t

Filter gain
-----------

The gain :math:`\\beta` represents all mean zero gyroscope measurement errors,
expressed as the magnitude of a quaternion derivative. It is defined using the
angular velocity:

.. math::
    \\beta = \\sqrt{\\frac{3}{4}}\\bar{\\omega}_\\beta

where :math:`\\bar{\\omega}_\\beta` is the estimated mean zero gyroscope
measurement error of each axis.

Footnotes
---------
.. [#] In reality, Earth's magnetic field varies slowly over time, which is a
    phenomenon known as `Geomagnetic secular variation <https://en.wikipedia.org/wiki/Geomagnetic_secular_variation>`_,
    but such shift can be omited for practical purposes.

"""

import numpy as np
from ..common.quaternion import Quaternion
from ..common.orientation import acc2q
from ..common.orientation import ecompass
from ..utils.core import _assert_numerical_iterable
from ..utils.core import _assert_numerical_positive_variable

class Madgwick:
    """
    Madgwick's Gradient Descent Orientation Filter

    If ``acc`` and ``gyr`` are given as parameters, the orientations will be
    immediately computed with method ``updateIMU``.

    If ``acc``, ``gyr`` and ``mag`` are given as parameters, the orientations
    will be immediately computed with method ``updateMARG``.

    Parameters
    ----------
    gyr : numpy.ndarray, default: None
        N-by-3 array with measurements of angular velocity in rad/s
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration in in m/s^2
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field in mT
    frequency : float, default: 100.0
        Sampling frequency in Herz.
    Dt : float, default: 0.01
        Sampling step in seconds. Inverse of sampling frequency. Not required
        if `frequency` value is given.
    gain : float, default: {0.033, 0.041}
        Filter gain. Defaults to 0.033 for IMU implementations, or to 0.041 for
        MARG implementations.
    gain_imu : float, default: 0.033
        Filter gain for IMU implementation.
    gain_marg : float, default: 0.041
        Filter gain for MARG implementation.
    q0 : numpy.ndarray, default: None
        Initial orientation, as a versor (normalized quaternion).

    Attributes
    ----------
    gyr : numpy.ndarray
        N-by-3 array with N tri-axial gyroscope samples.
    acc : numpy.ndarray
        N-by-3 array with N tri-axial accelerometer samples.
    mag : numpy.ndarray
        N-by-3 array with N tri-axial magnetometer samples.
    frequency : float
        Sampling frequency in Herz
    Dt : float
        Sampling step in seconds. Inverse of sampling frequency.
    gain : float
        Filter gain.
    q0 : numpy.ndarray
        Initial orientation, as a versor (normalized quaternion).

    Raises
    ------
    ValueError
        When dimension of input arrays ``acc``, ``gyr``, or ``mag`` are not
        equal.

    Examples
    --------

    This algorithm can work solely with gyroscope and accelerometer samples.
    The easiest way is to directly give the full array of samples to their
    matching parameters. The estimated quaternions are saved in the attribute
    ``Q``.

    >>> from ahrs.filters import Madgwick
    >>> madgwick = Madgwick(gyr=gyro_data, acc=acc_data)     # Using IMU
    >>> type(madgwick.Q), madgwick.Q.shape
    (<class 'numpy.ndarray'>, (1000, 4))

    If we desire to estimate each sample independently, we call the
    corresponding ``update`` method.

    >>> madgwick = Madgwick()
    >>> Q = np.tile([1., 0., 0., 0.], (num_samples, 1)) # Allocate for quaternions
    >>> for t in range(1, num_samples):
    ...     Q[t] = madgwick.updateIMU(Q[t-1], gyr=gyro_data[t], acc=acc_data[t])


    This algorithm requires a valid initial attitude, as a versor. This can be
    set with the parameter ``q0``:

    >>> madgwick = Madgwick(gyr=gyro_data, acc=acc_data, q0=[0.7071, 0.0, 0.7071, 0.0])

    .. warning::
        Do **NOT** initialize the filter with an empty array. The initial
        quaternion **must** be a versor, which is a quaternion, whose norm is
        equal to ``1.0``.

    If no initial orientation is given, an attitude is estimated using the
    first sample of each sensor. This initial attitude is computed assuming the
    sensors are straped to a body in a quasi-static state.

    Further on, we can also use magnetometer data.

    >>> madgwick = Madgwick(gyr=gyro_data, acc=acc_data, mag=mag_data)   # Using MARG

    A constant sampling frequency equal to 100 Hz is used by default. To change
    this value we set it in its parameter ``frequency``.

    >>> madgwick = Madgwick(gyr=gyro_data, acc=acc_data, frequency=150.0)   # 150 Hz

    Or, alternatively, setting the sampling step (:math:`\\Delta t = \\frac{1}{f}`):

    >>> madgwick = Madgwick(gyr=gyro_data, acc=acc_data, Dt=1/150)

    This is specially useful for situations where the sampling rate is variable:

    >>> madgwick = Madgwick()
    >>> Q = np.zeros((num_samples, 4))      # Allocation of quaternions
    >>> Q[0] = [1.0, 0.0, 0.0, 0.0]         # Initial attitude as a quaternion
    >>> for t in range(1, num_samples):
    >>>     madgwick.Dt = new_sample_rate
    ...     Q[t] = madgwick.updateIMU(Q[t-1], gyr=gyro_data[t], acc=acc_data[t])

    Madgwick's algorithm uses a gradient descent method to correct the
    estimation of the attitude. The **step size**, a.k.a.
    `learning rate <https://en.wikipedia.org/wiki/Learning_rate>`_, is
    considered a *gain* of this algorithm and can be set in the parameters too:

    >>> madgwick = Madgwick(gyr=gyro_data, acc=acc_data, gain=0.01)

    Following the original article, the gain defaults to ``0.033`` for IMU
    arrays, and to ``0.041`` for MARG arrays. Alternatively, the individual
    gains can be also set separately:

    >>> madgwick = Madgwick(gyr=gyro_data, acc=acc_data, mag=mag_data, gain_imu=0.01, gain_marg=0.05)

    """
    def __init__(self, gyr: np.ndarray = None, acc: np.ndarray = None, mag: np.ndarray = None, **kwargs):
        self.gyr: np.ndarray = gyr
        self.acc: np.ndarray = acc
        self.mag: np.ndarray = mag
        self.frequency: float = kwargs.get('frequency', 100.0)
        self.Dt: float = kwargs.get('Dt', (1.0/self.frequency) if self.frequency else 0.01)
        self.q0: np.ndarray = kwargs.get('q0')
        self._set_gain(**kwargs)
        self._assert_validity_of_inputs()
        if self.acc is not None and self.gyr is not None:
            self.Q: np.ndarray = self._compute_all()

    def _set_gain(self, **kwargs) -> None:
        """Set the gain parameter."""
        self.gain_imu: float = kwargs.get('gain_imu', 0.033)
        self.gain_marg: float = kwargs.get('gain_marg', 0.041)
        self.gain: float = kwargs.get('beta')  # Setting gain with `beta` will be removed in the future.
        if self.gain is None:
            self.gain: float = kwargs.get('gain', self.gain_imu if self.mag is None else self.gain_marg)

    def _assert_validity_of_inputs(self):
        """Asserts the validity of the inputs."""
        for item in ["frequency", "Dt", "gain", "gain_imu", "gain_marg"]:
            _assert_numerical_positive_variable(getattr(self, item), item)
        if self.q0 is not None:
            self.q0 = Quaternion(self.q0).to_array()

    def _compute_all(self) -> np.ndarray:
        """
        Estimate the quaternions given all data.

        Attributes ``gyr`` and ``acc`` must contain data. If ``mag`` contains
        data, the updateMARG() method is used.

        Returns
        -------
        Q : numpy.ndarray
            M-by-4 Array with all estimated quaternions, where M is the number
            of samples.

        """
        _assert_numerical_iterable(self.gyr, 'Angular velocity vector')
        _assert_numerical_iterable(self.acc, 'Gravitational acceleration vector')
        self.gyr = np.copy(self.gyr)
        self.acc = np.copy(self.acc)
        if self.acc.shape != self.gyr.shape:
            raise ValueError("acc and gyr are not the same size")
        num_samples = len(self.acc)
        Q = np.zeros((num_samples, 4))
        # Compute with IMU architecture
        if self.mag is None:
            Q[0] = acc2q(self.acc[0]) if self.q0 is None else self.q0/np.linalg.norm(self.q0)
            for t in range(1, num_samples):
                Q[t] = self.updateIMU(Q[t-1], self.gyr[t], self.acc[t])
            return Q
        # Compute with MARG architecture
        _assert_numerical_iterable(self.mag, 'Geomagnetic field vector')
        self.mag = np.copy(self.mag)
        if self.mag.shape != self.gyr.shape:
            raise ValueError("mag and gyr are not the same size")
        Q[0] = ecompass(self.acc[0], self.mag[0], frame='NED', representation='quaternion')
        for t in range(1, num_samples):
            Q[t] = self.updateMARG(Q[t-1], self.gyr[t], self.acc[t], self.mag[t])
        return Q

    def updateIMU(self, q: np.ndarray, gyr: np.ndarray, acc: np.ndarray, dt: float = None) -> np.ndarray:
        """
        Quaternion Estimation with IMU architecture.

        Parameters
        ----------
        q : numpy.ndarray
            A-priori quaternion.
        gyr : numpy.ndarray
            Sample of tri-axial Gyroscope in rad/s
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2
        dt : float, default: None
            Time step, in seconds, between consecutive Quaternions.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.

        Examples
        --------

        Assuming we have a tri-axial gyroscope array with 1000 samples, and
        1000 samples of a tri-axial accelerometer. We get the attitude with the
        Madgwick algorithm as:

        >>> from ahrs.filters import Madgwick
        >>> madgwick = Madgwick()
        >>> Q = np.tile([1., 0., 0., 0.], (len(gyro_data), 1)) # Allocate for quaternions
        >>> for t in range(1, num_samples):
        ...   Q[t] = madgwick.updateIMU(Q[t-1], gyr=gyro_data[t], acc=acc_data[t])
        ...

        Or giving the data directly in the class constructor will estimate all
        attitudes at once and store the estimated quaternions in the attribute
        ``Q``:

        >>> madgwick = Madgwick(gyr=gyro_data, acc=acc_data)
        >>> madgwick.Q.shape
        (1000, 4)

        """
        _assert_numerical_iterable(q, 'Quaternion')
        _assert_numerical_iterable(gyr, 'Tri-axial gyroscope sample')
        _assert_numerical_iterable(acc, 'Tri-axial accelerometer sample')
        q = Quaternion(q)
        dt = self.Dt if dt is None else dt
        if np.linalg.norm(gyr) == 0:
            return q.to_array()
        qDot = 0.5 * q.product([0, *gyr])                           # (eq. 12)
        a_norm = np.linalg.norm(acc)
        if a_norm > 0:
            a = acc/a_norm
            qw, qx, qy, qz = q/np.linalg.norm(q)
            # Objective function (eq. 25)
            f = np.array([2.0*(qx*qz - qw*qy)   - a[0],
                          2.0*(qw*qx + qy*qz)   - a[1],
                          2.0*(0.5-qx**2-qy**2) - a[2]])
            if np.linalg.norm(f) > 0:
                # Jacobian (eq. 26)
                J = np.array([[-2.0*qy,  2.0*qz, -2.0*qw, 2.0*qx],
                              [ 2.0*qx,  2.0*qw,  2.0*qz, 2.0*qy],
                              [ 0.0,    -4.0*qx, -4.0*qy, 0.0   ]])
                # Objective Function Gradient
                gradient = J.T@f                                    # (eq. 34)
                gradient /= np.linalg.norm(gradient)
                qDot -= self.gain*gradient                          # (eq. 33)
        q_new = q + qDot*dt                                         # (eq. 13)
        q_new /= np.linalg.norm(q_new)
        return q_new

    def updateMARG(self, q: np.ndarray, gyr: np.ndarray, acc: np.ndarray, mag: np.ndarray, dt: float = None) -> np.ndarray:
        """
        Quaternion Estimation with a MARG architecture.

        Parameters
        ----------
        q : numpy.ndarray
            A-priori quaternion.
        gyr : numpy.ndarray
            Sample of tri-axial Gyroscope in rad/s
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer in m/s^2
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer in nT
        dt : float, default: None
            Time step, in seconds, between consecutive Quaternions.

        Returns
        -------
        q : numpy.ndarray
            Estimated quaternion.

        Examples
        --------

        Assuming we have a tri-axial gyroscope array with 1000 samples, a
        second array with 1000 samples of a tri-axial accelerometer, and a
        third array with 1000 samples of a tri-axial magnetometer. We get the
        attitude with the Madgwick algorithm as:

        >>> from ahrs.filters import Madgwick
        >>> madgwick = Madgwick()
        >>> Q = np.tile([1., 0., 0., 0.], (len(gyro_data), 1)) # Allocate for quaternions
        >>> for t in range(1, num_samples):
        ...   Q[t] = madgwick.updateMARG(Q[t-1], gyr=gyro_data[t], acc=acc_data[t], mag=mag_data[t])
        ...

        Or giving the data directly in the class constructor will estimate all
        attitudes at once and store the estimated quaternions in the attribute
        ``Q``:

        >>> madgwick = Madgwick(gyr=gyro_data, acc=acc_data, mag=mag_data)
        >>> madgwick.Q.shape
        (1000, 4)

        """
        _assert_numerical_iterable(q, 'Quaternion')
        _assert_numerical_iterable(gyr, 'Tri-axial gyroscope sample')
        _assert_numerical_iterable(acc, 'Tri-axial accelerometer sample')
        _assert_numerical_iterable(mag, 'Tri-axial magnetometer sample')
        q = Quaternion(q)
        dt = self.Dt if dt is None else dt
        if np.linalg.norm(gyr) == 0:
            return q.to_array()
        if np.linalg.norm(mag) == 0:
            return self.updateIMU(q, gyr, acc)
        qDot = 0.5 * q.product([0, *gyr])                           # (eq. 12)
        a_norm = np.linalg.norm(acc)
        if a_norm > 0:
            a = acc/a_norm
            m = mag/np.linalg.norm(mag)
            # Rotate normalized magnetometer measurements
            q_m = Quaternion([0, *m])
            h = q.product(q_m.product(q.conj))                      # (eq. 45)
            bx = np.linalg.norm([h[1], h[2]])                       # (eq. 46)
            bz = h[3]
            qw, qx, qy, qz = q/np.linalg.norm(q)
            # Objective function (eq. 31)
            f = np.array([2.0*(qx*qz - qw*qy)   - a[0],
                          2.0*(qw*qx + qy*qz)   - a[1],
                          2.0*(0.5-qx**2-qy**2) - a[2],
                          2.0*bx*(0.5 - qy**2 - qz**2) + 2.0*bz*(qx*qz - qw*qy)       - m[0],
                          2.0*bx*(qx*qy - qw*qz)       + 2.0*bz*(qw*qx + qy*qz)       - m[1],
                          2.0*bx*(qw*qy + qx*qz)       + 2.0*bz*(0.5 - qx**2 - qy**2) - m[2]])
            if np.linalg.norm(f) > 0:
                # Jacobian (eq. 32)
                J = np.array([[-2.0*qy,               2.0*qz,              -2.0*qw,               2.0*qx             ],
                              [ 2.0*qx,               2.0*qw,               2.0*qz,               2.0*qy             ],
                              [ 0.0,                 -4.0*qx,              -4.0*qy,               0.0                ],
                              [-2.0*bz*qy,            2.0*bz*qz,           -4.0*bx*qy-2.0*bz*qw, -4.0*bx*qz+2.0*bz*qx],
                              [-2.0*bx*qz+2.0*bz*qx,  2.0*bx*qy+2.0*bz*qw,  2.0*bx*qx+2.0*bz*qz, -2.0*bx*qw+2.0*bz*qy],
                              [ 2.0*bx*qy,            2.0*bx*qz-4.0*bz*qx,  2.0*bx*qw-4.0*bz*qy,  2.0*bx*qx          ]])
                gradient = J.T@f                                    # (eq. 34)
                gradient /= np.linalg.norm(gradient)
                qDot -= self.gain*gradient                          # (eq. 33)
        q_new = q + qDot*dt                                         # (eq. 13)
        q_new /= np.linalg.norm(q_new)
        return q_new



# -*- coding: utf-8 -*-
"""
Implementation of the Fast Kalman Filter algorithm for orientation estimation
described in "Novel MARG-Sensor Orientation Algorithm Using Fast Kalman Filter"
:cite:p:`guo2017`.

The Fast Kalman Filter (FKF) is a quaternion-based orientation estimation
algorithm that focuses on the fusion of accelerometer and magnetometer
measurements.

In the FKF, the quaternion kinematic equation is the process model, similar to
the EKF. Likewise, the measurement model is based on the accelerometer and
magnetometer readings.

The FKF algorithm is computationally efficient, because it symbolically solves
the linear system of equations of the observation model.

The steps of the KF are summarized as:

1. **State Prediction**

.. math::
    \\mathbf{q}_t^- = \\Phi_t \\mathbf{q}_{t-1} + \\xi_t

2. **Covariance Prediction**

.. math::
    \\mathbf{\\Sigma}_{\\mathbf{q}_t^-} = \\Phi_t \\mathbf{\\Sigma}_{\\mathbf{q}_{t-1}} \\Phi_t^T + \\mathbf{\\Sigma}_{\\xi_t}

3. **Kalman Gain**

.. math::
    \\mathbf{G}_t = \\mathbf{\\Sigma}_{\\mathbf{q}_t^-} \\big [\\mathbf{\\Sigma}_{\\mathbf{q}_t^-} + \\mathbf{\\Sigma}_{v_t}\\big ]^{-1}

4. **State Update**

.. math::
    \\mathbf{q}_t = \\mathbf{q}_t + \\mathbf{G}_t (\\mathbf{y}_t - \\mathbf{q}_t)

5. **Covariance Update**

.. math::
    \\mathbf{\\Sigma}_{\\mathbf{q}_t} = (\\mathbf{I}_4 - \\mathbf{G}_t) \\mathbf{\\Sigma}_{\\mathbf{q}_t^-}

We update the state, :math:`\\mathbf{q}_t`, at every time :math:`t`, with the
predicted state :math:`\\mathbf{q}_t^-`, the measurements :math:`\\mathbf{y}_t`,
the state transition matrix :math:`\\Phi_t`, the covariance matrix
:math:`\\mathbf{P}_t`, the process noise covariance matrix
:math:`\\boldsymbol\\varepsilon_t`, and the Kalman Gain :math:`\\mathbf{G}_t`.

Prediction Model
----------------

The first two steps correspond to the prediction model, and are computed like
in the common `Extended Kalman Filter <ekf.html>`_.

The **state vector** is defined by the quaternion itself, without considering
biases over the sensors.

.. math::
    \\mathbf{q}_t = \\begin{bmatrix} q_w & q_x & q_y & q_z \\end{bmatrix}^T

The **state prediction** (we can call it the "predicted quaternion") is
estimated as:

.. math::
    \\boxed{\\mathbf{q}_t^- = \\Phi_t \\mathbf{q}_{t-1} + \\xi_t}

where :math:`\\xi_t` is the process noise, and the **state transition matrix**
:math:`\\Phi_t` is:

.. math::
    \\Phi_t = \\mathbf{I}_4 + \\frac{\\Delta t}{2} \\boldsymbol\\Omega(\\mathbf{\\omega}_t)

with the *omega operator* is defined as:

.. math::

    \\boldsymbol\\Omega(\\mathbf{\\omega}) =
    \\begin{bmatrix}
        0 & -\\mathbf{\\omega}^T \\\\
        \\mathbf{\\omega} & \\lfloor\\mathbf{\\omega}\\rfloor_\\times
    \\end{bmatrix} =
    \\begin{bmatrix}
        0 & -\\omega_x & -\\omega_y & -\\omega_z \\\\
        \\omega_x & 0 & \\omega_z & -\\omega_y \\\\
        \\omega_y & -\\omega_z & 0 & \\omega_x \\\\
        \\omega_z & \\omega_y & -\\omega_x & 0
    \\end{bmatrix}

We use the measured angular rates :math:`\\mathbf{\\omega}_t` from the
gyroscope to approximate the instantaneous change in the quaternion. Then, we
multiply by :math:`\\frac{\\Delta t}{2}` to numerically integrate it, and get
the predicted quaternion :math:`\\mathbf{q}_t^-`.

At step 2, the **Covariance Prediction** is computed with:

.. math::
    \\boxed{\\mathbf{\\Sigma}_{\\mathbf{q}_t^-} = \\Phi_t \\mathbf{\\Sigma}_{\\mathbf{q}_{t-1}} \\Phi_t^T + \\mathbf{\\Sigma}_{\\xi_t}}

where :math:`\\mathbf{\\Sigma}_{\\xi_t}` is the process noise covariance

.. math::

    \\mathbf{\\Sigma}_{\\xi_t} =
    \\Big(\\frac{\\Delta t}{2}\\Big)^2 \\mathbf{\\Xi}_t \\mathbf{\\Sigma}_{\\mathrm{gyro}} \\mathbf{\\Xi}_t^T

:math:`\\mathbf{\\Sigma}_{\\mathrm{gyro}}` is the angular rates' covariance,
usually defined as:

.. math::

    \\mathbf{\\Sigma}_{\\mathrm{gyro}} =
    \\begin{bmatrix}
        \\sigma_{\\omega_x}^2 & 0 & 0 \\\\
        0 & \\sigma_{\\omega_y}^2 & 0 \\\\
        0 & 0 & \\sigma_{\\omega_z}^2
    \\end{bmatrix}

The terms :math:`\\sigma_{\\omega_x}`, :math:`\\sigma_{\\omega_y}`, and
:math:`\\sigma_{\\omega_z}` are the standard deviations at each axis of the
gyroscope.

In this implementation, we assume they are equal, and we define
:math:`\\mathbf{\\Sigma}_{\\mathrm{gyro}}` as:

.. math::

    \\mathbf{\\Sigma}_{\\mathrm{gyro}} = \\sigma_{\\omega}^2 \\mathbf{I}_3 =
    \\begin{bmatrix}
        \\sigma_{\\omega}^2 & 0 & 0 \\\\
        0 & \\sigma_{\\omega}^2 & 0 \\\\
        0 & 0 & \\sigma_{\\omega}^2
    \\end{bmatrix}

:math:`\\mathbf{\\Xi}_t` is the matrix built from the quaternion's elements:

.. math::

    \\mathbf{\\Xi}_t =
    \\begin{bmatrix}
        q_x & q_y & q_z \\\\
        -q_w & -q_z & q_y \\\\
        q_z & -q_w & -q_x \\\\
        -q_y & q_x & -q_w
    \\end{bmatrix}

Correction Model
----------------

The last three steps are the correction model, where we update the predicted
state :math:`\\mathbf{q}_t^-` with the measurements :math:`\\mathbf{q}_{am}`.

The **Kalman Gain** is computed in step three as:

.. math::
    \\boxed{\\mathbf{G}_t = \\mathbf{\\Sigma}_{\\mathbf{q}_t^-} \\big [\\mathbf{\\Sigma}_{\\mathbf{q}_t^-} + \\mathbf{\\Sigma}_{v_t}\\big ]^{-1}}

where :math:`\\mathbf{\\Sigma}_{v_t}` is the measurement quaternion's
variance approximated with:

.. math::
    \\mathbf{\\Sigma}_{v_t} = \\mathbf{J} \\mathbf{\\Sigma}_{\\mathrm{am}} \\mathbf{J}^T

Defining :math:`\\mathbf{\\Sigma}_a` and :math:`\\mathbf{\\Sigma}_m` as the
covariances of accelerometer and magnetomer, respectively, we define the
covariance matrix :math:`\\mathbf{\\Sigma}_{\\mathrm{am}}` as:

.. math::

    \\mathbf{\\Sigma}_{\\mathrm{am}} =
    \\begin{bmatrix}
        \\mathbf{\\Sigma}_{\\mathbf{a}} & \\mathbf{0} \\\\
        \\mathbf{0} & \\mathbf{\\Sigma}_{\\mathbf{m}}
    \\end{bmatrix} =
    \\begin{bmatrix}
        \\sigma_{a_x}^2 & 0 & 0 & 0 & 0 & 0 \\\\
        0 & \\sigma_{a_y}^2 & 0 & 0 & 0 & 0 \\\\
        0 & 0 & \\sigma_{a_z}^2 & 0 & 0 & 0 \\\\
        0 & 0 & 0 & \\sigma_{m_x}^2 & 0 & 0 \\\\
        0 & 0 & 0 & 0 & \\sigma_{m_y}^2 & 0 \\\\
        0 & 0 & 0 & 0 & 0 & \\sigma_{m_z}^2
    \\end{bmatrix}

:math:`\\mathbf{J}` is the Jacobian matrix of the measurement model.

.. math::

    \\begin{array}{rcl}
    \\mathbf{J} &=& \\frac{\\partial \\mathbf{q}_{am}}{\\partial \\Big\\{ \\big(\\mathbf{A}^b\\big)^T \\, \\big(\\mathbf{M}^b\\big)^T\\Big\\} } \\\\
    &=& \\begin{bmatrix}
    \\frac{\\partial q_w}{\\partial a_x} & \\frac{\\partial q_w}{\\partial a_y} & \\frac{\\partial q_w}{\\partial a_z} &
    \\frac{\\partial q_w}{\\partial m_x} & \\frac{\\partial q_w}{\\partial m_y} & \\frac{\\partial q_w}{\\partial m_z} \\\\
    \\frac{\\partial q_x}{\\partial a_x} & \\frac{\\partial q_x}{\\partial a_y} & \\frac{\\partial q_x}{\\partial a_z} &
    \\frac{\\partial q_x}{\\partial m_x} & \\frac{\\partial q_x}{\\partial m_y} & \\frac{\\partial q_x}{\\partial m_z} \\\\
    \\frac{\\partial q_y}{\\partial a_x} & \\frac{\\partial q_y}{\\partial a_y} & \\frac{\\partial q_y}{\\partial a_z} &
    \\frac{\\partial q_y}{\\partial m_x} & \\frac{\\partial q_y}{\\partial m_y} & \\frac{\\partial q_y}{\\partial m_z} \\\\
    \\frac{\\partial q_z}{\\partial a_x} & \\frac{\\partial q_z}{\\partial a_y} & \\frac{\\partial q_z}{\\partial a_z} &
    \\frac{\\partial q_z}{\\partial m_x} & \\frac{\\partial q_z}{\\partial m_y} & \\frac{\\partial q_z}{\\partial m_z}
    \\end{bmatrix}
    \\end{array}

and the measurement quaternion :math:`\\mathbf{q}_{am}` is computed from the
accelerometer and magnetometer readings in the body frame.

.. math::

    \\mathbf{q}_{am} = \\frac{1}{4} \\Big(\\mathbf{W}_a + \\mathbf{I}_4\\Big) \\Big(\\mathbf{W}_m + \\mathbf{I}_4\\Big) \\mathbf{q}_{t-1}

with:

.. math::

    \\mathbf{W}_a = \\begin{bmatrix}
        a_z & a_y & -a_x & 0 \\\\
        a_y & -a_z & 0 & a_x \\\\
        -a_x & 0 & -a_z & a_y \\\\
        0 & a_x & a_y & a_z
    \\end{bmatrix}

and:

.. math::

    \\mathbf{W}_m = \\begin{bmatrix}
        m_z & m_y & -m_x & 0 \\\\
        m_y & -m_z & 0 & m_x \\\\
        -m_x & 0 & -m_z & m_y \\\\
        0 & m_x & m_y & m_z
    \\end{bmatrix}

What makes the FKF computationally efficient is that the Jacobian matrix
:math:`\\mathbf{J}`, and the measurement quaternion :math:`\\mathbf{q}_{am}`
are reduced symbolically, instead of numerically calculating them.

So, instead of performing several matrix multiplications and vector
operations, the FKF algorithm solves the linear system of equations
symbolically, and updates the state vector directly.

Then, we correct our predicted state :math:`\\mathbf{q}_t^-` with the
measurements :math:`\\mathbf{q}_{am}` using the Kalman Gain :math:`\\mathbf{G}_t`:

.. math::

    \\boxed{\\mathbf{q}_t = \\mathbf{q}_t^- + \\mathbf{G}_t (\\mathbf{q}_{am} - \\mathbf{q}_t^-)}

Finally, we update the covariance matrix:

.. math::

    \\boxed{\\mathbf{\\Sigma}_{\\mathbf{q}_t} = (\\mathbf{I}_4 - \\mathbf{G}_t) \\mathbf{\\Sigma}_{\\mathbf{q}_t^-}}

"""

from typing import Tuple
from typing import Optional
import numpy as np
from ..utils.core import _assert_numerical_iterable
from ..common.orientation import ecompass

class FKF:
    """
    Class of Fast Kalman Filter algorithm

    Parameters
    ----------
    gyr : numpy.ndarray, default: None
        N-by-3 array with measurements of angular velocity in rad/s
    acc : numpy.ndarray, default: None
        N-by-3 array with measurements of acceleration in in m/s^2
    mag : numpy.ndarray, default: None
        N-by-3 array with measurements of magnetic field in nT
    frequency : float, default: 100.0
        Sampling frequency in Hz.
    Dt : float, default: 0.01
        Sampling period in seconds. If ``frequency`` is given, ``Dt`` is
        computed as 1/frequency.
    sigma_g : float, default: 0.01
        Standard deviation of the gyroscope noise.
    sigma_a : float, default: 0.01
        Standard deviation of the accelerometer noise.
    sigma_m : float, default: 0.01
        Standard deviation of the magnetometer noise.
    Pk : float, default: 0.01
        Initial value of the diagonal values of covariance matrix.

    """
    def __init__(self, gyr: Optional[np.ndarray] = None, acc: Optional[np.ndarray] = None, mag: Optional[np.ndarray] = None, **kwargs):
        self.gyr: Optional[np.ndarray] = gyr
        self.acc: Optional[np.ndarray] = acc
        self.mag: Optional[np.ndarray] = mag
        self.frequency: float = kwargs.get('frequency', 100.0)
        self.Dt: float = kwargs.get('Dt', (1.0/self.frequency) if self.frequency else 0.01)
        self.sigma_g: float = kwargs.get('sigma_g', 0.01)
        self.sigma_a: float = kwargs.get('sigma_a', 0.01)
        self.sigma_m: float = kwargs.get('sigma_m', 0.01)
        self.Pk: np.ndarray = np.identity(4)*kwargs.get('Pk', 0.01)
        if all(x is not None for x in[self.gyr, self.acc, self.mag]):
            self.Q: np.ndarray = self._compute_all()

    def Omega4(self, x: np.ndarray) -> np.ndarray:
        """
        Omega operator.

        Given a vector :math:`\\mathbf{x}\\in\\mathbb{R}^3`, return a
        :math:`4\\times 4` matrix of the form:

        .. math::
            \\boldsymbol\\Omega(\\mathbf{x}) =
            \\begin{bmatrix}
            0 & -\\mathbf{x}^T \\\\ \\mathbf{x} & \\lfloor\\mathbf{x}\\rfloor_\\times
            \\end{bmatrix} =
            \\begin{bmatrix}
            0 & -x_1 & -x_2 & -x_3 \\\\
            x_1 & 0 & x_3 & -x_2 \\\\
            x_2 & -x_3 & 0 & x_1 \\\\
            x_3 & x_2 & -x_1 & 0
            \\end{bmatrix}

        Parameters
        ----------
        x : numpy.ndarray
            Three-dimensional vector.

        Returns
        -------
        Omega : numpy.ndarray
            Omega matrix.
        """
        return np.array([
            [0.0,  -x[0], -x[1], -x[2]],
            [x[0],   0.0,  x[2], -x[1]],
            [x[1], -x[2],   0.0,  x[0]],
            [x[2],  x[1], -x[0],   0.0]])

    def kalman_update(self, q_1: np.ndarray, q_am: np.ndarray, Pk_1: np.ndarray, Phi: np.ndarray, Sigma_eps: np.ndarray, Sigma_v: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Kalman update step.

        Parameters
        ----------
        q_1 : numpy.ndarray
            Previous state vector.
        q_am : numpy.ndarray
            Measurement vector.
        Pk_1 : numpy.ndarray
            Previous covariance matrix.
        Phi : numpy.ndarray
            State transition matrix.
        Sigma_eps : numpy.ndarray
            Process noise covariance matrix.
        Sigma_v : numpy.ndarray
            Measurement noise covariance matrix.

        Returns
        -------
        xk : numpy.ndarray
            Updated state vector.
        Pk : numpy.ndarray
            Updated covariance matrix.

        """
        # Prediction
        q_ = Phi @ q_1                                          # Predicted State
        Sigma_q_ = Phi @ (Pk_1 @ Phi.transpose()) + Sigma_eps   # Predicted Covariance
        # Update
        Gk = Sigma_q_ @ (np.linalg.inv(Sigma_q_ + Sigma_v))     # Kalman Gain
        Sigma_q = (np.identity(4) - Gk) @ Sigma_q_              # Updated Covariance
        q = q_ + Gk @ (q_am - q_)                               # Updated State
        return q, Sigma_q

    def measurement_quaternion_acc_mag(self, q: np.ndarray, acc: np.ndarray, mag: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Measurement model of Accelerometer and Magnetometer.

        Parameters
        ----------
        q : numpy.ndarray
            Predicted quaternion.
        acc : numpy.ndarray
            Sample of tri-axial Accelerometer.
        mag : numpy.ndarray
            Sample of tri-axial Magnetometer.

        Returns
        -------
        q : numpy.ndarray
            Measurement quaternion.

        """
        # Assert arrays and normalize observations
        _assert_numerical_iterable(q, 'Quaternion')
        _assert_numerical_iterable(acc, 'Tri-axial accelerometer sample')
        _assert_numerical_iterable(mag, 'Tri-axial magnetometer sample')
        ax, ay, az = acc / np.linalg.norm(acc)
        mx, my, mz = mag / np.linalg.norm(mag)
        qw, qx, qy, qz = q
        # Dynamic magnetometer reference vector (eq. 4)
        mD = ax*mx + ay*my + az*mz
        mN = np.sqrt(1.0 - mD**2)
        # Build measurement quaternion (eq. 25)
        q_new = np.zeros(4)
        q_new[0] =     (ay*mD*my + (1 + az)*(1 + mN*mx + mD*mz) + ax*(mD*mx - mN*mz))*qw              + ((mD + az*mD - ax*mN)*my + ay*(1 + mN*mx - mD*mz))*qx + (ay*mN*my + ax*(-1 + mN*mx + mD*mz) + (1 + az)*(-(mD*mx) + mN*mz))*qy                  + (-((ax*mD + mN + az*mN)*my) + ay*(mD*mx + mN*mz))*qz
        q_new[1] =     ((mD - az*mD - ax*mN)*my             + ay*(1 + mN*mx + mD*mz))*qw + (ay*mD*my - (-1 + az)*(1 + mN*mx - mD*mz) + ax*(mD*mx + mN*mz))*qx                  + ((ax*mD + mN - az*mN)*my + ay*(-(mD*mx) + mN*mz))*qy + (-(ay*mN*my) + ax*(1 - mN*mx + mD*mz) - (-1 + az)*(mD*mx + mN*mz))*qz
        q_new[2] = (-(ay*mN*my) - ax*(1 + mN*mx + mD*mz) + (-1 + az)*(mD*mx - mN*mz))*qw               + ((-(ax*mD) + mN - az*mN)*my + ay*(mD*mx + mN*mz))*qx   + (ay*mD*my + (-1 + az)*(-1 + mN*mx + mD*mz) + ax*(mD*mx - mN*mz))*qy                 + ((mD - az*mD + ax*mN)*my + ay*(1 - mN*mx + mD*mz))*qz
        q_new[3] = ax*(qx + mN*mx*qx + mN*my*qy + mN*mz*qz + mD*(my*qw - mz*qx + mx*qz)) + (1 + az)*(mD*mx*qx + mD*my*qy + qz + mD*mz*qz - mN*(my*qw - mz*qx + mx*qz)) + ay*(mN*mz*qw + mN*my*qx + qy - mN*mx*qy - mD*(mx*qw + mz*qy - my*qz))
        q_new = 0.25 * q_new / np.linalg.norm(q_new)
        # Build Jacobian matrix (eq. 27)
        J = np.zeros((4, 6))
        J[0, 0]= -qy - mN*(mz*qw + my*qx - mx*qy) + mD*(mx*qw + mz*qy - my*qz)
        J[0, 1]=  qx + mN*mx*qx + mN*my*qy + mN*mz*qz + mD*(my*qw - mz*qx + mx*qz)
        J[0, 2]=  qw + mN*mx*qw + mD*mz*qw + mD*my*qx - mD*mx*qy + mN*mz*qy - mN*my*qz
        J[0, 3]= (ax*mD + mN + az*mN)*qw + ay*mN*qx + (-((1 + az)*mD) + ax*mN)*qy + ay*mD*qz
        J[0, 4]= ay*mD*qw + (mD + az*mD - ax*mN)*qx + ay*mN*qy - (ax*mD + mN + az*mN)*qz
        J[0, 5]= mD*(qw + az*qw - ay*qx + ax*qy) + mN*(-(ax*qw) + qy + az*qy + ay*qz)
        J[1, 0]= qz - mN*(my*qw - mz*qx + mx*qz) + mD*(mx*qx + my*qy + mz*qz)
        J[1, 1]= qw + mN*mx*qw + mD*mz*qw + mD*my*qx - mD*mx*qy + mN*mz*qy - mN*my*qz
        J[1, 2]= -((1 + mN*mx)*qx) - mD*(my*qw - mz*qx + mx*qz) - mN*(my*qy + mz*qz)
        J[1, 3]= ay*(mN*qw - mD*qy) - (-1 + az)*(mN*qx + mD*qz) + ax*(mD*qx - mN*qz)
        J[1, 4]= mD*(qw - az*qw + ay*qx + ax*qy) - mN*(ax*qw + (-1 + az)*qy + ay*qz)
        J[1, 5]= ay*(mD*qw + mN*qy) + mD*((-1 + az)*qx + ax*qz) + mN*(ax*qx + qz - az*qz)
        J[2, 0]= -((1 + mN*mx + mD*mz)*qw) - mD*my*qx + mD*mx*qy - mN*mz*qy + mN*my*qz
        J[2, 1]= qz - mN*(my*qw - mz*qx + mx*qz) + mD*(mx*qx + my*qy + mz*qz)
        J[2, 2]= -qy - mN*(mz*qw + my*qx - mx*qy) + mD*(mx*qw + mz*qy - my*qz)
        J[2, 3]= mD*((-1 + az)*qw + ay*qx + ax*qy) - mN*(ax*qw + qy - az*qy + ay*qz)
        J[2, 4]= ay*(-(mN*qw) + mD*qy) - (-1 + az)*(mN*qx + mD*qz) + ax*(-(mD*qx) + mN*qz)
        J[2, 5]= mN*(qw - az*qw + ay*qx) - ax*(mD*qw + mN*qy) + mD*((-1 + az)*qy + ay*qz)
        J[3, 0]= qx + mN*mx*qx + mN*my*qy + mN*mz*qz + mD*(my*qw - mz*qx + mx*qz)
        J[3, 1]= qy + mN*(mz*qw + my*qx - mx*qy) - mD*(mx*qw + mz*qy - my*qz)
        J[3, 2]= qz - mN*(my*qw - mz*qx + mx*qz) + mD*(mx*qx + my*qy + mz*qz)
        J[3, 3]= -(ay*(mD*qw + mN*qy)) + ax*(mN*qx + mD*qz) + (1 + az)*(mD*qx - mN*qz)
        J[3, 4]= (1 + az)*(-(mN*qw) + mD*qy) + ax*(mD*qw + mN*qy) + ay*(mN*qx + mD*qz)
        J[3, 5]= ay*(mN*qw - mD*qy) + (1 + az)*(mN*qx + mD*qz) + ax*(-(mD*qx) + mN*qz)
        # J = 0.25 * J
        return q_new, J

    def _compute_all(self) -> np.ndarray:
        """
        Compute all steps of the FKF algorithm.

        Returns
        -------
        Q : numpy.ndarray
            Computed quaternions.

        """
        _assert_numerical_iterable(self.gyr, 'Angular velocity vector')
        _assert_numerical_iterable(self.acc, 'Gravitational acceleration vector')
        _assert_numerical_iterable(self.mag, 'Geomagnetic field vector')
        gyr = np.atleast_2d(self.gyr)
        acc = np.atleast_2d(self.acc)
        mag = np.atleast_2d(self.mag)
        if gyr.shape != acc.shape or gyr.shape != mag.shape:
            raise ValueError("All input arrays must have the same shape.")
        num_samples = gyr.shape[0]
        Sigma_g = self.sigma_g * np.identity(3)
        Sigma_am = np.diag([self.sigma_a]*3 + [self.sigma_m]*3)     # (eq. 28)
        Q = np.zeros((num_samples, 4))
        # Initial quaternion from the accelerometer and magnetometer
        Q[0] = ecompass(acc[0], mag[0], frame='NED', representation='quaternion')
        for t in range(1, num_samples):
            q_ = Q[t-1]                                             # Previous quaternion
            # PROCESS MODEL
            omega4 = self.Omega4(gyr[t])                            # Skew symmetric matrix (eq. 20)
            Phi = np.identity(4) + 0.5 * self.Dt * omega4           # State transition matrix (eq. 21)
            Xi = np.array([[ q_[1],  q_[2],  q_[3]],
                           [-q_[0], -q_[3], -q_[2]],
                           [ q_[2], -q_[0], -q_[1]],
                           [-q_[2],  q_[1], -q_[0]]])               # (eq. 24)
            Sigma_eps = (self.Dt/2.0)**2 * Xi @ Sigma_g @ Xi.transpose()   # Process noise covariance (eq. 23)
            # MEASUREMENT MODEL
            qy, J = self.measurement_quaternion_acc_mag(q_, acc[t], mag[t])
            Sigma_v = J @ Sigma_am @ J.transpose()                  # Measurement quaternion's covariance (eq. 26)
            # Kalman Update
            q, self.Pk = self.kalman_update(q_, qy, self.Pk, Phi, Sigma_eps, Sigma_v)
            Q[t] = q
        return Q
